{- 
  (Alternative to) Lazy I/O 
  Copyright 2013 Edsko de Vries, Well-Typed

  Pipes from scratch: Suspension
-}
{-# OPTIONS_GHC -Wall #-}

import Control.Monad.IO.Class
import Control.Applicative

newtype Pipe r = Pipe { unPipe :: IO (PipeStep r) }

data PipeStep r =
    Pure r
  | Suspend (Pipe r)

instance Monad Pipe where
  return x = Pipe $ return (Pure x)
  x >>= f  = Pipe $ do xstep <- unPipe x
                       case xstep of
                         Suspend k -> return $ Suspend (k >>= f)  
                         Pure r    -> unPipe (f r)
 
instance MonadIO Pipe where 
  liftIO io = Pipe $ Pure <$> io

suspend :: Pipe ()
suspend = Pipe . return $ Suspend (return ())

runPipe :: Pipe r -> IO r
runPipe p = do
  step <- unPipe p
  case step of
    Suspend k -> runPipe k 
    Pure r    -> return r

example1 :: Pipe ()
example1 = do
  liftIO $ putStr "Hello "
  suspend
  liftIO $ putStr "world\n"

example2 :: Pipe ()
example2 = liftIO $ putStr "crazy "

interleave :: Pipe r -> Pipe r -> Pipe r
interleave p q = Pipe $ do
  pstep <- unPipe p
  case pstep of 
    Suspend p' -> unPipe $ interleave q p' 
    Pure _a    -> unPipe $ q 

main :: IO ()
main = runPipe $ interleave example1 example2
