{- 
  (Alternative to) Lazy I/O 
  Copyright 2013 Edsko de Vries, Well-Typed

  Pipes from scratch: Leftovers in pipes style (ish) 
-}
{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE ScopedTypeVariables #-}

import Prelude hiding (dropWhile)
import Control.Monad
import Control.Monad.IO.Class
import Control.Applicative
import Data.Void -- from the 'void' package

--infixl 5 >->
--infixr 6 >~>

newtype Pipe l a b r = Pipe { unPipe :: [l] -> IO (PipeStep l a b r) }

data PipeStep l a b r =
    Pure [l] r
  | Request (a -> Pipe l a b r)
  | Respond b (Pipe l a b r) 

instance Monad (Pipe l a b) where 
  return x = Pipe $ \ls -> return (Pure ls x)
  x >>= f  = Pipe $ \ls -> do 
    xstep <- unPipe x ls
    case xstep of
      Request k   -> return $ Request (k >=> f)
      Respond b k -> return $ Respond b (k >>= f) 
      Pure ls' r  -> unPipe (f r) ls'
 
instance MonadIO (Pipe l a b) where 
  liftIO io = Pipe $ \ls -> Pure ls <$> io

respond :: b -> Pipe l a b ()
respond b = Pipe $ \_ -> return $ Respond b (return ())

request :: Pipe l r b r
request = Pipe $ \_ -> return $ Request return 

unDraw :: l -> Pipe l a b ()
unDraw l = Pipe $ \ls -> return $ Pure (l : ls) ()

draw :: Pipe r r b r
draw = Pipe $ \ls -> case ls of
                       []    -> return $ Request return
                       l:ls' -> return $ Pure ls' l

-- "Pull based composition"
(>->) :: forall l a b c r. Pipe l a b r -> Pipe l b c r -> Pipe l a c r
(>->) = goRight
  where
    goRight :: Pipe l a b r -> Pipe l b c r -> Pipe l a c r
    goRight p q = Pipe $ \ls -> do
      qstep <- unPipe q ls
      case qstep of
        Respond b k -> return $ Respond b (goRight p k)
        Pure ls' r  -> return $ Pure ls' r
        Request k   -> unPipe (goLeft p k) ls

    goLeft :: Pipe l a b r -> (b -> Pipe l b c r) -> Pipe l a c r
    goLeft p q = Pipe $ \ls -> do
      pstep <- unPipe p ls
      case pstep of
        Request k   -> return $ Request (\a -> goLeft (k a) q)
        Pure ls' r  -> return $ Pure ls' r
        Respond b k -> unPipe (goRight k (q b)) ls

runPipe :: Pipe l () Void r -> IO r 
runPipe p = do
  step <- unPipe p []
  case step of
    Pure _ r    -> return r
    Request k   -> runPipe (k ()) 
    Respond b _ -> absurd b

dropWhile :: (l -> Bool) -> Pipe l l b ()
dropWhile p = do
  x <- request
  if p x
    then dropWhile p
    else unDraw x

idPipe :: Pipe a a a r
idPipe = forever $ draw >>= respond 

prompter :: Pipe l a Int r 
prompter = do 
  liftIO $ putStrLn "Hi!"
  forever $ do
    i <- liftIO $ putStr "> " >> readLn
    respond i

printer :: Pipe l Int b r
printer = forever $ do
  i <- request
  liftIO $ print i 

main1 :: IO ()
main1 = runPipe (prompter >-> (dropWhile even >> idPipe) >-> printer)

