{- 
  (Alternative to) Lazy I/O 
  Copyright 2013 Edsko de Vries, Well-Typed

  Pipes from scratch: Pipe as a monad transformer 
-}
{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE ScopedTypeVariables #-}

import Control.Monad
import Control.Monad.IO.Class
import Control.Applicative
import Data.Void -- from the 'void' package
import Control.Monad.Trans.Class

infixl 5 >->

newtype Pipe a b m r = Pipe { unPipe :: m (PipeStep a b m r) }

data PipeStep a b m r =
    Pure r
  | Request (a -> Pipe a b m r)
  | Respond b (Pipe a b m r) 

instance Monad m => Monad (Pipe a b m) where
  return x = Pipe $ return (Pure x)
  x >>= f  = Pipe $ do xstep <- unPipe x
                       case xstep of
                         Request k   -> return $ Request (k >=> f)
                         Respond b k -> return $ Respond b (k >>= f)
                         Pure r      -> unPipe (f r)

instance MonadTrans (Pipe a b) where
  lift p = Pipe $ Pure `liftM` p 

instance MonadIO m => MonadIO (Pipe a b m) where 
  liftIO = lift . liftIO 

respond :: Monad m => b -> Pipe a b m ()
respond b = Pipe . return $ Respond b (return ())

request :: Monad m => Pipe r b m r
request = Pipe . return $ Request return 

-- "Pull based composition"
(>->) :: forall a b c m r. Monad m => Pipe a b m r -> Pipe b c m r -> Pipe a c m r
(>->) = goRight
  where
    goRight :: Pipe a b m r -> Pipe b c m r -> Pipe a c m r
    goRight p q = Pipe $ do
      qstep <- unPipe q
      case qstep of
        Respond b k -> return $ Respond b (goRight p k)
        Pure r      -> return $ Pure r
        Request k   -> unPipe $ goLeft p k

    goLeft :: Pipe a b m r -> (b -> Pipe b c m r) -> Pipe a c m r
    goLeft p q = Pipe $ do
      pstep <- unPipe p
      case pstep of
        Request k   -> return $ Request (\a -> goLeft (k a) q)
        Pure r      -> return $ Pure r
        Respond b k -> unPipe $ goRight k (q b)
                      
runPipe :: Monad m => Pipe () Void m r -> m r 
runPipe p = do
  step <- unPipe p
  case step of
    Pure r      -> return r
    Request k   -> runPipe (k ()) 
    Respond b _ -> absurd b

prompter :: MonadIO m => Pipe a Int m r 
prompter = do 
  liftIO $ putStrLn "Hi!"
  forever $ do
    i <- liftIO $ putStr "> " >> readLn
    respond i

printer :: MonadIO m => Pipe Int b m r
printer = forever $ do
  i <- request
  liftIO $ print i 

main1 :: IO ()
main1 = runPipe (prompter >-> printer)
